"""
Provides :class:`Config` and some related helpers.
"""

from __future__ import annotations

__author__ = "Patrick Doetsch"
__credits__ = ["Patrick Doetsch", "Paul Voigtlaender"]

from typing import TypeVar, Type, Optional, Union, Any, Dict, List, Sequence, Tuple
import contextlib
import sys
import os
import types as _types


T = TypeVar("T")


class Config:
    """
    Reads in some config file, and provides access to the key/value items.
    We support some simple text-line-based config, JSON, and Python format.
    """

    def __init__(self, items: Optional[Dict[str, Any]] = None):
        """
        :param items: optional initial typed_dict
        """
        self.dict: Dict[str, List[str]] = {}
        self.typed_dict: Dict[str, Any] = {}  # could be loaded via JSON or so
        self.network_topology_json: Optional[str] = None
        self.files = []
        if items is not None:
            self.typed_dict.update(items)

    def __getstate__(self):
        import io
        from returnn.util.task_system import Pickler

        class _CustomPickler(Pickler):
            use_whichmodule = False

            # This will trigger potentially some fallback of task_system.Pickler,
            # like _save_type_fallback.
            # Also, do not use the intelligent logic for our own dict.
            # We explicitly want to pickle it as-is.
            module_name_black_list = {_PyModuleName}

        buffer = io.BytesIO()
        pickler = _CustomPickler(buffer)
        memo_idx = len(pickler.memo)
        pickler.memo[id(self)] = memo_idx, self
        pickler.dump(self.typed_dict)

        return {
            "_pid": os.getpid(),
            "_self_memo_idx": memo_idx,
            "_typed_dict_pickled": buffer.getvalue(),
            "_is_global": self is get_global_config(raise_exception=False),
        }

    def __setstate__(self, state):
        import io

        # Use pure-Python unpickling to be able to extend the memo.
        # noinspection PyUnresolvedReferences,PyProtectedMember
        from pickle import _Unpickler

        self.__init__()

        buffer = io.BytesIO(state["_typed_dict_pickled"])
        unpickler = _Unpickler(buffer)
        unpickler.memo[state["_self_memo_idx"]] = self
        self.typed_dict = unpickler.load()

        if state["_is_global"] and os.getpid() != state["_pid"]:
            set_global_config(self)
            _global_config_as_py_module_proxy_setup()

    def load_file(self, f):
        """
        Reads the configuration parameters from a file and adds them to the inner set of parameters.

        :param string|io.TextIOBase|io.StringIO f:
        """
        if isinstance(f, str):
            assert os.path.isfile(f), "config file not found: %r" % f
            self.files.append(f)
            filename = f
            dirname = os.path.dirname(filename) or "."
            content = open(filename).read()
        else:
            # assume stream-like
            filename = "<config string>"
            dirname = None
            content = f.read()
        content = content.strip()
        if content.startswith("#!") or filename.endswith(".py"):  # assume Python
            if dirname and os.path.exists(f"{dirname}/__init__.py") and filename.endswith(".py"):
                # It looks like a Python module inside a Python package.
                # Import it as a module.
                import importlib

                basedir = os.path.abspath(dirname)
                while os.path.exists(f"{basedir}/__init__.py"):
                    basedir = os.path.dirname(basedir)
                if basedir not in sys.path:
                    sys.path.insert(0, basedir)
                modname = os.path.relpath(dirname, basedir).replace("/", ".") + "." + os.path.basename(filename)[:-3]
                mod = importlib.import_module(modname)
                self.update(vars(mod))

            else:
                # Directly execute the Python code.
                from returnn.util.basic import custom_exec

                # Operate inplace on ourselves.
                # Also, we want that it's available as the globals() dict, so that defined functions behave well
                # (they would lose the local context otherwise).
                user_ns = self.typed_dict
                # Always overwrite:
                user_ns.update({"config": self, "__file__": filename, "__name__": _PyModuleName})
                custom_exec(content, filename, user_ns, user_ns)
                _global_config_as_py_module_proxy_setup()
            return
        if content.startswith("{"):  # assume JSON
            from returnn.util.basic import load_json

            json_content = load_json(content=content)
            assert isinstance(json_content, dict)
            self.update(json_content)
            return
        # old line-based format
        for line in content.splitlines():
            if "#" in line:  # Strip away comment.
                line = line[: line.index("#")]
            line = line.strip()
            if not line:
                continue
            line = line.split(None, 1)
            assert len(line) == 2, "unable to parse config line: %r" % line
            self.add_line(key=line[0], value=line[1])

    @classmethod
    def get_config_file_type(cls, f: str) -> str:
        """
        :param f: file path
        :return: "py", "js" or "txt"
        """
        with open(f, "r") as f:
            start = f.read(3)
        if start.startswith("#!"):
            return "py"
        if start.startswith("{"):
            return "js"
        return "txt"

    def parse_cmd_args(self, args: Sequence[str]):
        """
        :param args:
        """
        from optparse import OptionParser

        parser = OptionParser()
        parser.add_option(
            "-a",
            "--activation",
            dest="activation",
            help="[STRING/LIST] Activation functions: logistic, tanh, softsign, relu, identity, zero, one, maxout.",
        )
        parser.add_option(
            "-b",
            "--batch_size",
            dest="batch_size",
            help="[INTEGER/TUPLE] Maximal number of frames per batch (optional: shift of batching window).",
        )
        parser.add_option(
            "-c",
            "--chunking",
            dest="chunking",
            help="[INTEGER/TUPLE] Maximal number of frames per sequence (optional: shift of chunking window).",
        )
        parser.add_option("-d", "--description", dest="description", help="[STRING] Description of experiment.")
        parser.add_option("-e", "--epoch", dest="epoch", help="[INTEGER] Starting epoch.")
        parser.add_option("-E", "--eval", dest="eval", help="[STRING] eval file path")
        parser.add_option(
            "-f",
            "--gate_factors",
            dest="gate_factors",
            help="[none/local/global] Enables pooled (local) or separate (global) coefficients on gates.",
        )
        parser.add_option("-g", "--lreg", dest="lreg", help="[FLOAT] L1 or L2 regularization.")
        parser.add_option(
            "-i",
            "--save_interval",
            dest="save_interval",
            help="[INTEGER] Number of epochs until a new model will be saved.",
        )
        parser.add_option("-j", "--dropout", dest="dropout", help="[FLOAT] Dropout probability (0 to disable).")
        parser.add_option(
            "-k", "--output_file", dest="output_file", help="[STRING] Path to target file for network output."
        )
        parser.add_option("-l", "--log", dest="log", help="[STRING] Log file path.")
        parser.add_option("-L", "--load", dest="load", help="[STRING] load model file path.")
        parser.add_option(
            "-m", "--momentum", dest="momentum", help="[FLOAT] Momentum term in gradient descent optimization."
        )
        parser.add_option(
            "-n", "--num_epochs", dest="num_epochs", help="[INTEGER] Number of epochs that should be trained."
        )
        parser.add_option("-o", "--order", dest="order", help="[default/sorted/random] Ordering of sequences.")
        parser.add_option("-p", "--loss", dest="loss", help="[loglik/sse/ctc] Objective function to be optimized.")
        parser.add_option(
            "-q",
            "--cache",
            dest="cache",
            help="[INTEGER] Cache size in bytes (supports notation for kilo (K), mega (M) and gigabyte (G)).",
        )
        parser.add_option(
            "-r",
            "--learning_rate",
            dest="learning_rate",
            help="[FLOAT] Learning rate in gradient descent optimization.",
        )
        parser.add_option(
            "-s", "--hidden_sizes", dest="hidden_sizes", help="[INTEGER/LIST] Number of units in hidden layers."
        )
        parser.add_option(
            "-t",
            "--truncate",
            dest="truncate",
            help="[INTEGER] Truncates sequence in BPTT routine after specified number of timesteps (-1 to disable).",
        )
        parser.add_option(
            "-u",
            "--device",
            dest="device",
            help="[STRING/LIST] CPU and GPU devices that should be used (example: gpu0,cpu[1-6] or gpu,cpu*).",
        )
        parser.add_option("-v", "--verbose", dest="log_verbosity", help="[INTEGER] Verbosity level from 0 - 5.")
        parser.add_option("-w", "--window", dest="window", help="[INTEGER] Width of sliding window over sequence.")
        parser.add_option("-x", "--task", dest="task", help="[train/forward/analyze] Task of the current program call.")
        parser.add_option(
            "-y", "--hidden_type", dest="hidden_type", help="[VALUE/LIST] Hidden layer types: forward, recurrent, lstm."
        )
        parser.add_option(
            "-z", "--max_sequences", dest="max_seqs", help="[INTEGER] Maximal number of sequences per batch."
        )
        parser.add_option("--config", dest="load_config", help="[STRING] load config")
        (options, args) = parser.parse_args(list(args))
        options = vars(options)
        for opt in options.keys():
            if options[opt] is not None:
                if opt == "load_config":
                    self.load_file(options[opt])
                else:
                    self.add_line(opt, options[opt])
        assert len(args) % 2 == 0, "expect (++key, value) config tuples in remaining args: %r" % args
        for i in range(0, len(args), 2):
            key, value = args[i : i + 2]
            assert key[0:2] == "++", "expect key prefixed with '++' in (%r, %r)" % (key, value)
            if value[:2] == "+-":
                value = value[1:]  # otherwise we never could specify things like "++threshold -0.1"
            self.add_line(key=key[2:], value=value)

    def add_line(self, key: str, value: str):
        """
        Adds one specific configuration (key,value) pair to the inner set of parameters

        :param key:
        :param value:
        """
        if key in self.typed_dict:
            # This is a special case. We overwrite a config value which was typed before.
            # E.g. this could have been loaded via a Python config file.
            # We want to keep the entry in self.typed_dict because there might be functions/lambdas inside
            # the config which require the global variable to be available.
            # See :func:`test_rnn_init_config_py_global_var`.
            value_type = type(self.typed_dict[key])
            if value_type == str:
                pass  # keep as-is
            else:
                try:
                    value = eval(value)
                except SyntaxError:
                    from returnn.log import log

                    print(
                        "WARNING: can't evaluate config param %r to previous type: %s. Keeping as string."
                        % (value, value_type),
                        file=log.v1,
                    )
            self.typed_dict[key] = value
            return
        if value.find(",") > 0:
            value = value.split(",")
        else:
            value = [value]
        if key == "include":
            for f in value:
                self.load_file(f)
        else:
            self.dict[key] = value

    def has(self, key: str) -> bool:
        """
        Returns whether the given key is present in the inner set of parameters

        :param key:
        :returns: True if and only if the given key is in the inner set of parameters
        """
        if key in self.typed_dict:
            return True
        return key in self.dict

    def is_typed(self, key: str) -> bool:
        """
        :param key:
        :returns: True if and only if the value of the given key has a specified data type
        """
        return key in self.typed_dict

    def is_true(self, key: str, default: bool = False) -> bool:
        """
        :param key:
        :param default:
        :return: bool(value) if it is set or default
        """
        if self.is_typed(key):
            return bool(self.typed_dict[key])
        return self.bool(key, default=default)

    def is_of_type(self, key: str, types: Union[type, Tuple[type, ...]]) -> bool:
        """
        :param key:
        :param types: for isinstance() check
        :return: whether is_typed(key) is True and isinstance(value, types) is True
        """
        if key in self.typed_dict:
            return isinstance(self.typed_dict[key], types)
        return False

    def get_of_type(self, key: str, types: Union[type, Tuple[type, ...], Type[T]], default: Optional[T] = None) -> T:
        """
        :param str key:
        :param types: for isinstance() check
        :param default:
        :return: if is_of_type(key, types) is True, returns the value, otherwise default
        """
        if self.is_of_type(key, types):
            return self.typed_dict[key]
        return default

    def set(self, key: str, value: Any):
        """
        :param key:
        :param value:
        """
        self.typed_dict[key] = value

    def update(self, dikt: Dict[str, Any]):
        """
        :param dikt: dict
        """
        for key, value in dikt.items():
            self.set(key, value)

    def _hack_value_reading_debug(self):
        orig_value_func = self.value

        def wrapped_value_func(*args, **kwargs):
            """
            Wrapped func.
            """
            res = orig_value_func(*args, **kwargs)
            print(
                "Config.value(%s) -> %r"
                % (", ".join(list(map(repr, args)) + ["%s=%r" % (k, v) for (k, v) in kwargs.items()]), res)
            )
            return res

        setattr(self, "value", wrapped_value_func)

    def value(self, key: str, default: T, index: Optional[int] = None, list_join_str: str = ",") -> Union[T, str]:
        """
        :param key:
        :param default:
        :param index:
        :param list_join_str:
        """
        if key in self.typed_dict:
            ls = self.typed_dict[key]
            if index is None:
                if isinstance(ls, (list, tuple)):
                    return list_join_str.join([str(v) for v in ls])
                elif ls is None:
                    return default
                else:
                    return str(ls)
            else:
                return str(ls[index])
        if key in self.dict:
            ls = self.dict[key]
            if index is None:
                return list_join_str.join(ls)
            else:
                return ls[index]
        return default

    def typed_value(self, key: str, default: Optional[T] = None, index: Optional[int] = None) -> Union[T, Any]:
        """
        :param key:
        :param default:
        :param index:
        """
        value = self.typed_dict.get(key, default)
        if index is not None:
            assert isinstance(index, int)
            if isinstance(value, (list, tuple)):
                value = value[index]
            else:
                assert index == 0
        return value

    def opt_typed_value(self, key: str, default: Optional[T] = None) -> Union[T, Any, str]:
        """
        :param key:
        :param default:
        """
        if key in self.typed_dict:
            return self.typed_dict[key]
        return self.value(key, default)

    def int(self, key: str, default: T, index: int = 0) -> Union[int, T]:
        """
        Parses the value of the given key as integer, returning default if not existent

        :param key:
        :param default:
        :param index:
        """
        if key in self.typed_dict:
            value = self.typed_value(key, default=default, index=index)
            if value is not None:
                assert isinstance(value, int)
            return value
        if key in self.dict:
            return int(self.value(key, default, index))
        return default

    def bool(self, key: str, default: T, index: int = 0) -> Union[bool, T]:
        """
        Parses the value of the given key as boolean, returning default if not existent

        :param key:
        :param default:
        :param index:
        """
        if key in self.typed_dict:
            value = self.typed_value(key, default=default, index=index)
            if isinstance(value, int):
                value = bool(value)
            if value is not None:
                assert isinstance(value, bool)
            return value
        if key not in self.dict:
            return default
        v = str(self.value(key, None, index))
        if not v:
            return default
        from returnn.util.basic import to_bool

        return to_bool(v)

    def bool_or_other(self, key: str, default: Optional[T] = None, index: int = 0) -> Union[bool, T, Any]:
        """
        :param key:
        :param default:
        :param index:
        :return: if we have typed value, just as-is. otherwise try to convert to bool. or default if not there.
        """
        if key in self.typed_dict:
            return self.typed_value(key, default=default, index=index)
        if key not in self.dict:
            return default
        v = str(self.value(key, None, index))
        if not v:
            return default
        from returnn.util.basic import to_bool

        try:
            return to_bool(v)
        except ValueError:
            return v

    def float(self, key: str, default: T, index: int = 0) -> Union[float, T]:
        """
        Parses the value of the given key as float, returning default if not existent

        :param key:
        :param default:
        :param index:
        """
        if key in self.typed_dict:
            value = self.typed_value(key, default=default, index=index)
        else:
            value = self.value(key, default, index)
        if value is not None:
            if isinstance(value, str):
                # Special case for float as str. We automatically cast this case.
                # This is also to handle special values such as "inf".
                value = float(value)
            assert isinstance(value, (int, float))
        return value

    def list(self, key: str, default: Optional[T] = None) -> Union[List[str], T]:
        """
        :param key:
        :param default:
        """
        if default is None:
            default = []
        if key in self.typed_dict:
            value = self.typed_value(key, default=default)
            if value is None:
                return default
            if not isinstance(value, (tuple, list)):
                value = [value]
            return list(value)
        if key not in self.dict:
            return default
        return self.dict[key]

    def int_list(self, key: str, default: Optional[T] = None) -> Union[List[int], T]:
        """
        :param key:
        :param default:
        """
        if default is None:
            default = []
        if key in self.typed_dict:
            value = self.typed_value(key, default=default)
            if value is None:
                return default
            if not isinstance(value, (tuple, list)):
                value = [value]
            for x in value:
                assert isinstance(x, int)
            return list(value)
        return [int(x) for x in self.list(key, default)]

    def float_list(self, key: str, default: Optional[T] = None) -> Union[List[float], T]:
        """
        :param key:
        :param default:
        """
        if default is None:
            default = []
        if key in self.typed_dict:
            value = self.typed_value(key, default=default)
            if value is None:
                return default
            if not isinstance(value, (tuple, list)):
                value = [value]
            for x in value:
                assert isinstance(x, (float, int))
            return list(value)
        return [float(x) for x in self.list(key, default)]

    def int_pair(self, key: str, default: Optional[Tuple[int, int]] = None) -> Tuple[int, int]:
        """
        :param key:
        :param default:
        """
        if default is None:
            default = (0, 0)
        if not self.has(key):
            return default
        if key in self.typed_dict:
            value = self.typed_value(key, default=default)
            if not isinstance(value, (tuple, list)):
                value = (value, value)
            assert len(value) == 2
            for x in value:
                assert isinstance(x, int)
            return tuple(value)
        value = self.value(key, "")
        if ":" in value:
            return int(value.split(":")[0]), int(value.split(":")[1])
        else:
            return int(value), int(value)


_global_config: Optional[Config] = None


@contextlib.contextmanager
def global_config_ctx(config: Optional[Config]):
    """
    sets the config as global config in this context,
    and recovers the original global config afterwards
    """
    global _global_config
    prev_global_config = _global_config
    try:
        set_global_config(config)
        yield
    finally:
        _global_config = prev_global_config


def set_global_config(config: Optional[Config]):
    """
    Will define the global config, returned by :func:`get_global_config`

    :param config:
    """
    _get_or_set_config_via_tf_default_graph(config, assign=True)
    global _global_config
    _global_config = config


def get_global_config(*, raise_exception: bool = True, auto_create: bool = False, return_empty_if_none: bool = False):
    """
    :param raise_exception: if no global config is found, raise an exception, otherwise return None
    :param auto_create: if no global config is found, it creates one, registers it as global, and returns it
    :param return_empty_if_none: if no global config is found, it creates one (which is empty) and returns it
    :rtype: Config|None
    """
    config = _get_or_set_config_via_tf_default_graph()
    if config:
        return config
    if _global_config:
        return _global_config
    # We are the main process.
    import sys

    main_mod = sys.modules["__main__"]  # should be rnn.py
    if hasattr(main_mod, "config") and isinstance(main_mod.config, Config):
        return main_mod.config
    # Maybe __main__ is not rnn.py, or config not yet loaded.
    # Anyway, try directly. (E.g. for SprintInterface.)
    import returnn.__main__ as rnn

    if isinstance(rnn.config, Config):
        return rnn.config
    if auto_create:
        config = Config()
        set_global_config(config)
        return config
    if return_empty_if_none:
        return Config()
    if raise_exception:
        raise Exception("No global config found.")
    return None


def _get_or_set_config_via_tf_default_graph(config: Optional[Config] = None, assign: bool = False):
    """
    This is done in a safe way, and might just be a no-op.
    When TF is not imported yet, it will just return.

    :param config: if set, will set it
    :param assign: whether to set it (even if config=None)
    :rtype: Config|None
    """
    if "tensorflow" not in sys.modules:
        return None
    from returnn.tf.compat import v1 as tf_v1

    graph = tf_v1.get_default_graph()
    # We could use collection refs, but this could cause other problems,
    # and is more complicated than what we need.
    # We just use a custom own attrib.
    attrib_name = "_RETURNN_config_in_graph"
    if assign or config:
        setattr(graph, attrib_name, config)
    return getattr(graph, attrib_name, None)


def network_json_from_config(config):
    """
    :param Config config:
    :rtype: dict[str]
    """
    if config.has("network") and config.is_typed("network"):
        json_content = config.typed_value("network")
        assert isinstance(json_content, dict)
        assert json_content
        return json_content
    else:
        raise ValueError("Network is not defined in config. Define `network`.")


def tf_should_use_gpu(config):
    """
    :param Config config:
    :rtype: bool
    """
    cfg_dev = config.value("device", None)
    # Short path.
    if cfg_dev == "gpu":
        return True
    if cfg_dev == "cpu":
        return False
    if not cfg_dev:
        # Better default: Use GPU if available.
        from returnn.log import log
        from returnn.tf.util.basic import is_gpu_available

        if is_gpu_available():
            print("Device not set explicitly, and we found a GPU, which we will use.", file=log.v2)
            config.set("device", "gpu")
            return True
        else:
            print("Device not set explicitly, and no GPU found.", file=log.v2)
            config.set("device", "cpu")
            return False
    else:
        raise ValueError("Currently unsupported TF device %r specified" % (cfg_dev,))


_PyModuleName = "__returnn_config__"


def _global_config_as_py_module_proxy_setup():
    if _PyModuleName in sys.modules:
        return
    sys.modules[_PyModuleName] = _GlobalConfigAsPyModuleProxy(_PyModuleName)


class _GlobalConfigAsPyModuleProxy(_types.ModuleType):
    """
    Takes :func:`get_global_config`, and makes its ``typed_dict`` available as module attributes.
    """

    @staticmethod
    def _get_config() -> Optional[Config]:
        """
        :return: config or None if not available anymore
        """
        return get_global_config(raise_exception=False)

    def __getattribute__(self, item):
        # Implement also __getattribute__ such that early access to just self.__dict__ (e.g. via vars(self)) also works.
        if item == "__dict__":
            cfg: Optional[Config] = self._get_config()
            if not cfg:
                return {}
            return cfg
        return super().__getattribute__(item)

    def __getattr__(self, item):
        cfg: Optional[Config] = self._get_config()
        if not cfg:
            raise AttributeError("config %s not loaded anymore" % self.__name__)
        if item not in cfg.typed_dict:
            raise AttributeError("config %s has no attribute %r" % (self.__name__, item))
        return cfg.typed_dict[item]

    def __dir__(self):
        cfg: Optional[Config] = self._get_config()
        if not cfg:
            return []
        return sorted(cfg.typed_dict.keys())

    def __setattr__(self, key, value):
        if key in ["__file__"]:
            super().__setattr__(key, value)
            return
        cfg: Optional[Config] = self._get_config()
        if cfg:
            cfg.typed_dict[key] = value


class SubProcCopyGlobalConfigPreInitFunc:
    """
    A pre-init hook for a subprocess which copies the global config to the subprocess.

    It can be important that this init function is called even before the unpickling of other data,
    as that unpickling might depend on the right context, e.g. having the global RETURNN config.
    Example issue with MultiProcDataset: https://github.com/rwth-i6/returnn/issues/1495

    Example usage::

        NonDaemonicSpawnContext(process_pre_init_func=SubProcCopyGlobalConfigPreInitFunc())
    """

    def __init__(self):
        # Get the RETURNN global config here. Allow this to be optional (for use outside of RETURNN).
        # We store it here such that pickling this worker init func will also pickle the config,
        # so that we can reset it as global config inside the worker.
        # Some RETURNN datasets might depend on the config.
        # https://github.com/rwth-i6/returnn/issues/1495
        # MultiProcDataset has a similar logic, see https://github.com/rwth-i6/returnn/issues/1384.
        self.global_config = get_global_config(raise_exception=False)

    def __call__(self):
        from returnn.util import better_exchook
        from returnn.log import log
        from returnn import __old_mod_loader__

        better_exchook.install()
        __old_mod_loader__.disable_lazy_mod_loads()

        if self.global_config:
            set_global_config(self.global_config)
            log.init_by_config(self.global_config)
        else:
            if not log.initialized:
                log.initialize(verbosity=[5])



"""
Loop. Provides :class:`Loop`.
"""

from __future__ import annotations
from typing import Dict, Any, Optional, List, Union, Iterable
import tree
from returnn.util.basic import NotSpecified
from returnn.tensor import Tensor, Dim, ControlFlowContext
import returnn.frontend as rf
import returnn.tf.frontend_layers as rfl
from . import _utils


__all__ = ["Loop", "LoopModule"]


class Loop:
    """
    This represents a RecLayer subnetwork in RETURNN,
    i.e. where the calculation per step is defined explicitly.

    (For RecLayer with a predefined unit, see :class:`Rec`.
     Or for example :class:`Lstm`.)

    To define a loop like this pseudo Python code::

      x  # given, shape (batch, time, dim)
      h = Zeros([batch,dim])()  # initial state, shape (batch,dim)
      out = []
      for t in range(x.max_seq_len):
        x_lin = Linear(dim)(x[t])
        h_prev = h
        h = Linear(dim)(x_lin + h_prev)
        out.append(h)

      h  # final state
      out  # shape (time, batch, h_dim)

    You would write::

      dim = nn.FeatureDim(...)
      loop = nn.Loop(axis=...)
      loop.state.h = nn.zeros([batch_dim,dim])  # initial state
      with loop:
        x_t = loop.unstack(x)
        x_lin = Linear(dim)(x_t)
        loop.state.h = Linear(dim)(x_lin + loop.state.h)
        out = loop.stack(loop.state.h)

    ``state`` is :class:`Loop._StateHolder` and manages the recurrent state.

    This code must be run within a :func:`Module.forward`
    or with some active global name context (:class:`NameCtx`).

    This API is currently in development, and might change.
    See: https://github.com/rwth-i6/returnn_common/issues/16
    """

    def __init__(
        self,
        *,
        max_seq_len: Optional[Tensor] = NotSpecified,
        optimize_move_layers_out: Optional[bool] = NotSpecified,
        unroll: bool = NotSpecified,
        axis: Optional[Dim] = NotSpecified,
        debug: Optional[bool] = NotSpecified,
        name: str = "loop",
    ):
        super(Loop, self).__init__()
        self._has_given_axis = True
        if not axis or axis is NotSpecified:
            self._has_given_axis = False
            axis = Dim(None, name=f"{name}-dim")
        assert isinstance(axis, Dim)
        self.extra_opts = {
            {"max_seq_len": "max_seq_len_via"}.get(key, key): value
            for (key, value) in locals().items()
            if value is not NotSpecified and value is not None and key not in {"self", "__class__", "name"}
        }
        self.layer_module = LoopModule(loop=self)
        parent_name_ctx = rfl.Layer.current_ctx()
        self.control_flow_ctx = ControlFlowContext(
            kind=ControlFlowContext.Types.Loop,
            outer_ctx=rfl.Layer.inner_control_flow(),
            identifier=parent_name_ctx.get_abs_name(),
        )
        self.control_flow_ctx.loop_spatial_dim = axis
        self.name_ctx = rfl.Layer(
            module=self.layer_module,
            suggested_name=name,
            parent=parent_name_ctx,
            new_control_flow_ctx=self.control_flow_ctx,
            can_access_children=False,
        )
        self.name_ctx.custom_layer_name_scope = ""
        self.name_ctx.is_subnet = True
        self.name_ctx.extend_reserved_names({"output", "end"})
        self._entered_scope = False
        self._exited_scope = False
        self._exited_scope_with_exception = False
        self._state = _LoopStateHolder(loop=self)
        self.unstacked_refs = []  # type: List[Tensor]
        self.outputs = []  # type: List[Tensor]
        self._last_frames = {}  # type: Dict[rfl.Layer, Tensor]  # inner name -> outer
        self.axis = axis
        self.end_ref = None  # type: Optional[Tensor]
        self._iter_idx_ref = None  # type: Optional[Tensor]

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name_ctx.get_abs_name_repr()}>"

    def __enter__(self) -> Loop:
        assert not self._entered_scope, f"{self}: cannot enter twice"
        self._entered_scope = True
        self.name_ctx.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert not self._exited_scope, f"{self}: cannot exit twice"
        try:
            if not exc_type:
                if self.end_ref is None and not self.unstacked_refs:
                    raise Exception(f"{self}: call `unstack` or `end` at least once to define the loop length")
                # Make sure there is an "output" layer. (Similar as for Module with subnetwork.)
                if not self.outputs:
                    # Stack some dummy output
                    if self.end_ref is not None:
                        self.stack(self.end_ref)
                    else:
                        assert self.unstacked_refs
                        self.stack(self.unstacked_refs[0])
                assert "output" in self.name_ctx.children
        finally:
            self._exited_scope_with_exception = bool(exc_type)
            self._exited_scope = True
            self.name_ctx.__exit__(exc_type, exc_val, exc_tb)
        if not exc_type:
            res = self.layer_module()  # create the rec layer itself
            if self.end_ref is not None:
                res.raw_tensor.layer_extra_dependencies.append(self.end_ref.raw_tensor)

    @property
    def has_entered_scope(self) -> bool:
        """
        :return: whether we have entered the scope, i.e. we define the per-step calculation.
        """
        return self._entered_scope

    @property
    def state(self) -> Union[_LoopStateHolder, rf.State]:
        """state holder inside the loop"""
        if not self._exited_scope:
            return self._state
        if self._exited_scope_with_exception:  # nicer for debugging
            return self._state
        # noinspection PyProtectedMember
        return self._state._get_last()

    @state.setter
    def state(self, initial_state: rf.State):
        assert len(self._state) == 0, f"can only assign {self}.state once for the initial state"
        assert not self._entered_scope
        for key, value in initial_state.items():
            self._state[key] = value

    def unstack(self, source: Tensor, *, name: Optional[str] = None) -> Tensor:
        """
        Unrolls over the specified axis, and provides each frame in each loop iteration.
        The axis can be specified globally for the :class:`Loop` instance (recommended)
        or locally here (not recommended).
        """
        assert self._has_given_axis, "%s: unstack() requires a given axis" % self
        assert self.axis in source.dims
        res = _rec_unstack(source, axis=self.axis, name=name)
        self.unstacked_refs.append(res)
        return res

    def stack(self, source: Tensor, *, name: Optional[str] = None) -> Tensor:
        """
        Accumulates the frames of source within the loop,
        to make it accessible outside the loop.
        """
        # We don't need to do anything special because RETURNN RecLayer will automatically accumulate the frames
        # when we marked a layer with is_output_layer, and we access it from outside the loop.
        if not name and "output" not in self.name_ctx.children:
            name = self.name_ctx.get_child("output")
        if isinstance(name, str) or not name:
            if not name:
                name = source.name
            name = rfl.Layer(suggested_name=name, parent=self.name_ctx)
        assert isinstance(name, rfl.Layer)

        # We access the returned layer-ref from outside, thus fix the data template.
        data = source.copy_template().copy_add_dim_by_tag(dim_tag=self.axis, unbroadcast=True, axis=0)
        data.time_dim_axis = 0
        data.control_flow_ctx = self.control_flow_ctx.outer_ctx
        # Because the self.axis dim is yet unknown,
        # the logic of _register_dim_deps_when_novel will create a length layer
        # to make the length available.
        # However, this length layer will be inside the rec loop, which we do not want...
        with self.name_ctx.parent:
            res = rfl.make_layer({"class": "copy", "from": source}, predefined_out_data=data, name=name)
        assert isinstance(res, Tensor)
        if res.raw_tensor.name != "output":
            res.raw_tensor.layer_dict["is_output_layer"] = True
        self.outputs.append(res)
        return res

    def last(self, source: Tensor, *, name: Optional[str] = None) -> Tensor:
        """
        Gets the last value from source.
        """
        assert isinstance(source, Tensor)
        if source.raw_tensor in self._last_frames:
            return self._last_frames[source.raw_tensor]
        assert self.name_ctx.tensor is not None, f"{self}.last(...): must call from outside"  # current restriction...
        # need_last option only works in root of subnet of RecLayer
        if source.raw_tensor.parent is not self.name_ctx:
            assert self.name_ctx in source.raw_tensor.get_abs_name_ctx_list(), f"invalid {self}.last({source})"
            sub_layer_name = source.raw_tensor.get_name_in_ctx(self.name_ctx).replace("/", ".")
            source = _utils.copy(source, name=self.name_ctx.get_new_child(sub_layer_name))
            assert source.raw_tensor.parent is self.name_ctx
        source.raw_tensor.layer_dict["need_last"] = True
        sub_layer_name = source.raw_tensor.get_name_in_ctx(self.name_ctx)
        with self.name_ctx.parent:  # need to be outside the loop
            res = rfl.make_layer(
                {"class": "rec_last_output", "rec_layer": self.name_ctx.tensor, "sub_layer_name": sub_layer_name},
                predefined_out_data=source,
                name=name or sub_layer_name.replace("/", "_"),
            )
            res.raw_tensor.tensor_remove_unused_cleanup_hooks.append(
                lambda _: source.raw_tensor.layer_dict.pop("need_last")
            )
            res.raw_tensor.layer_extra_dependencies.append(source.raw_tensor)
            source.raw_tensor.usages.append(res.raw_tensor)
            self._last_frames[source.raw_tensor] = res
            return res

    def end(self, source: Tensor, *, include_eos: bool) -> Tensor:
        """
        For loops with dynamic ending condition (which might not use unstack),
        this defines the ending condition.

        :param source: the ending condition
        :param include_eos: if True, the last() and stack() function include the current ending frame, otherwise not
        """
        assert not self.end_ref, f"{self}.end() can only be called once"
        assert source.dtype == "bool", f"{self}: end expects boolean condition, got {source}"
        if self.axis.dyn_size_ext is None:
            dyn_size_ext = source.copy_template()
            dyn_size_ext.dtype = "int32"
            if dyn_size_ext.control_flow_ctx:
                dyn_size_ext.control_flow_ctx = dyn_size_ext.control_flow_ctx.outer_ctx
            self.axis.dyn_size_ext = dyn_size_ext
            self.axis.batch = dyn_size_ext.batch
            self.axis.control_flow_ctx = dyn_size_ext.control_flow_ctx
        self.extra_opts["include_eos"] = include_eos
        self.end_ref = _utils.copy(source, name=self.name_ctx.get_child("end"))
        return self.end_ref

    @property
    def max_seq_len(self) -> Optional[Tensor]:
        """max seq length in case the length is dynamic via :func:`end`"""
        return self.extra_opts.get("max_seq_len_via")

    @max_seq_len.setter
    def max_seq_len(self, value: Optional[Tensor]):
        if value is None:
            self.extra_opts.pop("max_seq_len_via", None)
        else:
            self.extra_opts["max_seq_len_via"] = value

    @property
    def iter_idx(self) -> Tensor:
        """
        The index of the current iteration, inside the loop. This is a scalar. This always starts with 0.

        """
        assert self._entered_scope and not self._exited_scope
        if self._iter_idx_ref is not None:
            return self._iter_idx_ref
        self._iter_idx_ref = self.name_ctx.get_child_tensor(
            ":i",
            data=Tensor(":i", dtype="int32", dim_tags=(), sparse_dim=self.axis, control_flow_ctx=self.control_flow_ctx),
        )
        return self._iter_idx_ref


class LoopModule(rf.Module):
    """
    This module is used internally by :class:`Loop` to create the RETURNN :class:`RecLayer` for the loop.
    This module would not be directly used by the user.
    """

    def __init__(self, loop: Loop):
        super(LoopModule, self).__init__()
        self.loop = loop

    def __call__(self) -> Tensor:
        """
        Makes layer dict for this loop, i.e. a RecLayer.
        """
        name_ctx = self.loop.name_ctx
        out = name_ctx.children["output"].tensor
        # self.stack already added the loop.axis dim tag to prepare the access from outside the loop.
        assert out.dim_tags[0] == self.loop.axis
        return rfl.make_layer(
            {"class": "rec", "from": [], "unit": name_ctx.make_net(), **self.loop.extra_opts},
            name=name_ctx,
            predefined_out_data=out,
        )


class _LoopStateHolder:
    def __init__(self, loop: Loop):
        self._loop = loop
        self._state = {}  # type: Dict[str, _LoopState]

    def __repr__(self):
        return f"{self._loop}.state"

    def _get_state(self, name: str) -> _LoopState:
        if name in self._state:
            return self._state[name]
        raise AttributeError(f"{self}: Unknown state attrib {name!r}. Assign the initial state first.")

    def _get_last(self) -> rf.State:
        return rf.State({key: value.get_last() for (key, value) in self._state.items()})

    def __getitem__(self, item):
        return self._get_state(item).get()

    def __setitem__(self, key, value):
        if not self._loop.has_entered_scope:
            assert key not in self._state, f"{self} already has state {key!r}"
            self._state[key] = _LoopState(name=key, loop=self._loop, initial=value)
            return
        self._get_state(key).assign(value)

    def __getattr__(self, item):
        return self[item]

    def __setattr__(self, key, value):
        if key in {"_state", "_loop"}:
            return super().__setattr__(key, value)
        self[key] = value

    def keys(self) -> Iterable[str]:
        """keys"""
        return self._state.keys()

    def values(self) -> List[Any]:
        """values"""
        return [v.get() for v in self._state.values()]

    def __len__(self):
        return len(self._state)

    def deep_tensors(self) -> List[Tensor]:
        """See :func:`LayerState.cls_deep_tensors`."""
        return rf.State.cls_deep_tensors(self)


class _LoopState:
    """
    Represents some recurrent state, to be used with :class:`Loop`.
    It can also represent some nested hierarchy of states.
    """

    def __init__(self, *, name: str, loop: Loop, initial: Union[Tensor, Any]):
        """
        :param name:
        :param loop:
        :param initial: some layer-ref, or any kind of nested structure of layers.
        """
        super(_LoopState, self).__init__()
        assert initial is not None
        initial = tree.map_structure(rf.convert_to_tensor, initial)
        self.initial = initial
        self.loop = loop
        self.name = name
        self.assigned_value = None
        self.name_ctx = tree.map_structure_with_path(
            lambda path, ref: rfl.Layer(
                suggested_name=".".join(str(key) for key in ("state", name) + path), parent=loop.name_ctx
            ),
            self.initial,
        )

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r}>"

    def assign(self, value):
        """
        Assign the new value of the current iteration.
        This is called (only) inside the loop.
        This will define the value for the next iteration.
        """
        assert self.name_ctx is not None
        assert value is not None
        assert self.assigned_value is None, (
            f"Cannot assign the rec state {self.loop}/{self.name} multiple times, "
            f"assigned previously to {self.assigned_value}, now to {value}"
        )
        tree.assert_same_structure(self.initial, value)
        tree.assert_same_structure(self.name_ctx, value)
        self.assigned_value = value

        def _map_ref_to_name_ctx(tensor: Tensor, name_ctx: rfl.Layer, initial: Tensor):
            assert isinstance(tensor, Tensor)
            assert isinstance(name_ctx, rfl.Layer)
            assert name_ctx.tensor is None, f"Loop state {name_ctx} already assigned"

            tensor.raw_tensor.make_all_sub_networks_and_optimize()

            layer_ctx_list = tensor.raw_tensor.get_abs_name_ctx_list()
            assert (
                self.loop.name_ctx in layer_ctx_list
            ), f"Loop state {name_ctx} should get a value inside the loop but got {tensor}"
            # We need some special logic for MaskedComputation but maybe also for others later.
            # This is currently not nice, but I'm not sure about better solutions.
            for i in range(layer_ctx_list.index(self.loop.name_ctx) + 1, len(layer_ctx_list) - 1):
                ctx, ctx_ = layer_ctx_list[i : i + 2]
                assert isinstance(ctx, rfl.Layer) and isinstance(ctx_, rfl.Layer)
                if isinstance(ctx.module, rfl.MaskedComputationModule):
                    ctx_.layer_dict["is_output_layer"] = True
                    break

            # Potential optimization for RETURNN layers.
            # See ReturnnWrappedLayerBase._get_recurrent_state.
            if tensor.raw_tensor.layer_dict:
                _do_const_initial_value_opt = False
                _const_initial_value_opt_layer_white_list = {"cum_concat", "rec"}
                if tensor.raw_tensor.layer_dict["class"] in _const_initial_value_opt_layer_white_list:
                    _do_const_initial_value_opt = True
                elif tensor.raw_tensor.layer_dict["class"] == "get_last_hidden_state":
                    src = tensor.raw_tensor.layer_dict["from"]
                    assert isinstance(src, Tensor)
                    if src.raw_tensor.layer_dict:
                        if src.raw_tensor.layer_dict["class"] in _const_initial_value_opt_layer_white_list:
                            _do_const_initial_value_opt = True
                if _do_const_initial_value_opt:
                    # Note: Only do this optimization for some layers because otherwise
                    # we might rely on the initial output shape.
                    initial_const = _utils.constant_value(initial)
                    if initial_const is not None:
                        initial = initial_const

                if tensor.raw_tensor.layer_dict["class"] == "get_last_hidden_state":
                    used_state_eliminate_optimization = False
                    key = tensor.raw_tensor.layer_dict.get("key", "state")
                    src = tensor.raw_tensor.layer_dict["from"]
                    assert isinstance(src, Tensor)
                    src_state_opt = src.raw_tensor.layer_dict.get("state") if src.raw_tensor.layer_dict else None
                    if isinstance(src_state_opt, rf.State):
                        src_state_for_key = src_state_opt.get(key)
                        if isinstance(src_state_for_key, rfl.PrevTensorRef):
                            if src_state_for_key.cur_layer_name_ctx is name_ctx:
                                # The 'state' argument of the rec layer refers to "prev:..." of the state.
                                # So we don't need to pass it now.
                                used_state_eliminate_optimization = True
                                src_state_opt[key] = None
                                if all(opt is None for opt in tree.flatten(src_state_opt)):
                                    del src.raw_tensor.layer_dict["state"]
                                # We need to pass the initial_state instead though.
                                src_initial_state_opt = src.raw_tensor.layer_dict.setdefault(
                                    "initial_state", rf.State()
                                )
                                src_initial_state_opt[key] = initial
                                # If there is any other code which refers to this state, it can access the passed layer.
                                # So anyway pass through.

                    if not used_state_eliminate_optimization:
                        raise NotImplementedError(
                            f"{self}.assign to {tensor} on {src}:"
                            f" We need https://github.com/rwth-i6/returnn_common/issues/31"
                            f" and https://github.com/rwth-i6/returnn/issues/732."
                        )

                else:  # class != get_last_hidden_state
                    if tensor.raw_tensor.layer_dict["class"] == "cum_concat":
                        layer_state_opt = tensor.raw_tensor.layer_dict.get("state")
                        if isinstance(layer_state_opt, rf.State) and set(layer_state_opt.keys()) == {"state"}:
                            layer_state = layer_state_opt.state
                            if (
                                isinstance(layer_state, rfl.PrevTensorRef)
                                and layer_state.cur_layer_name_ctx is name_ctx
                            ):
                                # The 'state' argument refers to "prev:..." of itself.
                                # This is redundant, so we don't need to pass it.
                                tensor.raw_tensor.layer_dict.pop("state")

                    assert "initial_state" not in tensor.raw_tensor.layer_dict
                    assert "initial_output" not in tensor.raw_tensor.layer_dict
                    tensor.raw_tensor.layer_dict["initial_output"] = initial

            else:  # tensor not Tensor
                raise NotImplementedError(f"{self}.assign to {tensor} (type {type(tensor)}) but Tensor expected")

            # Note: We assume this has been used before in get() -> PrevTensorRef.get_prev_ref().
            prev_name_ctx = name_ctx.parent.children.get(f"prev:{name_ctx.name}")
            if prev_name_ctx:  # might not exist if we have never accessed the prev state
                prev_ref = prev_name_ctx.tensor
                assert isinstance(prev_ref, rfl.PrevTensorRef), f"{name_ctx, prev_name_ctx}"
                if tensor.raw_tensor.parent != self.loop.name_ctx:
                    # Currently, RETURNN does not properly support a state in a subnet.
                    # So we copy the layer to the loop root under the reserved existing name.
                    _utils.copy(tensor, name=name_ctx)
                    if tensor.raw_tensor.layer_dict:
                        assert "initial_state" not in tensor.raw_tensor.layer_dict  # not supported/implemented
                        if "initial_output" in tensor.raw_tensor.layer_dict:
                            name_ctx.layer_dict["initial_output"] = tensor.raw_tensor.layer_dict.pop("initial_output")
                else:
                    prev_ref.assign_new_cur_tensor_name_ctx(tensor.raw_tensor)

            return tensor.raw_tensor

        self.name_ctx = tree.map_structure(_map_ref_to_name_ctx, value, self.name_ctx, self.initial)

    @staticmethod
    def _map_name_ctx_to_prev_tensor(name_ctx: rfl.Layer, initial: Tensor) -> rfl.PrevTensorRef:
        assert isinstance(name_ctx, rfl.Layer)
        return rfl.PrevTensorRef.get_prev_ref(cur_layer_name_ctx=name_ctx, initial=initial)

    def get(self):
        """
        Return prev or current value of the current loop iteration,
        depending on whether assign() already has been called or not.
        This is called (only) inside a loop.
        """
        assert self.name_ctx is not None
        if not self.loop.has_entered_scope:
            return self.initial
        if self.assigned_value is None:  # not yet assigned
            # Return prev value
            return tree.map_structure(self._map_name_ctx_to_prev_tensor, self.name_ctx, self.initial)
        return self.assigned_value

    def _map_name_ctx_to_last_tensor(self, name_ctx: rfl.Layer) -> Tensor:
        assert isinstance(name_ctx, rfl.Layer)
        assert name_ctx.tensor is not None, f"{self.loop} state {name_ctx} not assigned?"
        assert self.loop.name_ctx.tensor is not None, f"{self.loop} not yet exited?"
        return self.loop.last(name_ctx.tensor)

    def get_last(self):
        """
        Outside the loop, get the last instance.
        """
        assert self.name_ctx is not None
        assert self.assigned_value is not None
        return tree.map_structure(self._map_name_ctx_to_last_tensor, self.name_ctx)


# noinspection PyShadowingBuiltins,PyShadowingNames
def _rec_unstack(
    source: Tensor,
    *,
    axis: Dim,
    declare_rec_time: bool = NotSpecified,
    name: Optional[Union[str, rfl.Layer]] = None,
) -> Tensor:
    """
    This is supposed to be used inside a :class:`RecLayer`.
    The input is supposed to be outside the rec layer (i.e. via ``base:``).
    Uses tf.TensorArray and then unstack on the inputs to make it available per-frame.
    This is an alternative to making some input to the rec layer,
    such that the rec layer can have multiple inputs (as long as they have the same time dim).

    Note that due to automatic optimization, this layer will be optimized out of the rec loop anyway,
    and then the tf.TensorArray logic happens internally in RecLayer,
    thus we do not need to care about this here.
    (See get_input_moved_out for some internal handling.)

    Effectively, this layer is very similar to :class:`CopyLayer`,
    with the only special behavior that it checks (or even assigns) the loop dimension of RecLayer.

    Due to automatic optimization, not much happens here.
    The real logic happens in :func:`get_out_data_from_opts`.

    Note that it is allowed to leave both `axis` and `declare_rec_time` unset,
    in case you assign `axis` to the rec layer, and the source here has the same axis (dim tag).

    :param nn.Tensor source:
    :param nn.Dim axis:
    :param bool declare_rec_time:
    :param str|nn.NameCtx|None name:
    :return: layer
    """
    if not isinstance(source, Tensor):
        raise TypeError(f"rec_unstack: unexpected type for source {source!r}, need tensor")
    args = {
        "axis": axis,
        "declare_rec_time": declare_rec_time,
    }
    args = {key: value for (key, value) in args.items() if value is not NotSpecified}
    return rfl.make_layer({"class": "rec_unstack", "from": source, **args}, name=name or "rec_unstack")



"""
Python extension module compiler
"""

from __future__ import annotations

import sys
import sysconfig
from .native_code_compiler import NativeCodeCompiler


class PyExtModCompiler(NativeCodeCompiler):
    """
    Python extension module compiler
    """

    CacheDirName = "returnn_py_ext_mod_cache"

    def __init__(self, include_paths=(), **kwargs):
        py_compile_vars = sysconfig.get_config_vars()
        include_paths = list(include_paths) + [py_compile_vars["INCLUDEPY"]]
        super().__init__(
            include_paths=include_paths,
            **kwargs,
        )
        self._py_compile_vars = py_compile_vars
        self._py_mod = None

    _relevant_info_keys = NativeCodeCompiler._relevant_info_keys + ("py_version",)

    def _extra_common_opts(self):
        base_flags = super()._extra_common_opts()
        py_compile_flags = self._py_compile_vars["CFLAGS"].split() if self._py_compile_vars["CFLAGS"] else []
        return base_flags + py_compile_flags

    def _make_info_dict(self):
        d = super()._make_info_dict()
        d.update({"py_version": sys.version_info[:2]})
        return d

    def load_py_module(self):
        """
        :return: Python extension module
        """
        from importlib.util import spec_from_loader, module_from_spec
        from importlib.machinery import ExtensionFileLoader

        if self._py_mod:
            return self._py_mod

        self._maybe_compile()
        mod_name = self.base_name

        spec = spec_from_loader(mod_name, ExtensionFileLoader(mod_name, self._so_filename))  # noqa
        mod = module_from_spec(spec)
        spec.loader.exec_module(mod)  # noqa
        self._py_mod = mod
        return mod



"""
Benchmark for RETURNN frontend (RF) with PyTorch (PT) backend.

You can run this file directly (then it will just call `returnn.__main__.main`)
or use as a RETURNN config file.

This trains a standard AED model
on some dummy data.
"""

from __future__ import annotations

import subprocess
from typing import Optional, Any, Tuple, Dict, Sequence, List
import os
import sys
import tree
import numpy
import argparse
import tempfile
import time
import multiprocessing

if __name__ == "__main__":
    import _setup_returnn_env  # noqa

from returnn.tensor import Tensor, TensorDict, Dim, single_step_dim, batch_dim
import returnn.frontend as rf
from returnn.frontend.tensor_array import TensorArray
from returnn.frontend.encoder.conformer import ConformerEncoder, ConformerConvSubsample
from returnn import __main__

_my_file = os.path.abspath(__file__)
_my_dir = os.path.dirname(_my_file)
_returnn_root_dir = os.path.dirname(_my_dir)

config = dict(
    task="train",
    backend="torch",
    torch_dataloader_opts=dict(num_workers=1),
    # batching="laplace:.1000",  -- dummy dataset does not support this
    batch_size=15_000,
    max_seqs=200,
    max_seq_length_default_target=75,
    accum_grad_multiple_step=2,
    # gradient_clip=0,
    # gradient_clip_global_norm = 1.0
    optimizer={"class": "nadam", "epsilon": 1e-8},
    # gradient_noise=0.0,
    learning_rate=0.0005,
    learning_rates=(
        # matching pretraining
        list(numpy.linspace(0.0001, 0.001, num=10)) * 3
        + list(numpy.linspace(0.0001, 0.0005, num=10))
        + [0.0005] * 20
        + list(numpy.linspace(0.0005, 0.001, num=20))
    ),
    min_learning_rate=0.001 / 50,
    learning_rate_control="newbob_multi_epoch",
    learning_rate_control_relative_error_relative_lr=True,
    relative_error_div_by_old=True,
    use_learning_rate_control_always=True,
    newbob_multi_update_interval=1,
    learning_rate_control_min_num_epochs_per_new_lr=1,
    learning_rate_decay=0.9,
    newbob_relative_error_threshold=-0.01,
    num_epochs=1000,
    log_verbosity=[5],
)


# depends on dataset
extern_data_inputs_name = "data"
extern_data_targets_name = "classes"
data_spatial_dim = Dim(None, name="in_spatial")
data_feature_dim = Dim(2, name="in_feature")
targets_spatial_dim = Dim(None, name="out_spatial")
targets_dim = Dim(10, name="vocab")
extern_data = {
    extern_data_inputs_name: dict(
        dims=[
            batch_dim,
            data_spatial_dim,
        ],
        sparse_dim=data_feature_dim,
        dtype="int32",
        available_for_inference=True,
    ),
    extern_data_targets_name: dict(
        dims=[
            batch_dim,
            targets_spatial_dim,
        ],
        sparse_dim=targets_dim,
        dtype="int32",
        available_for_inference=False,
    ),
}


def _get_dataset_opts(name: str):
    opts = {
        "class": "TaskNumberBaseConvertDataset",
        "input_base": data_feature_dim.dimension,
        "output_base": targets_dim.dimension,
    }
    if name == "train":
        opts["num_seqs"] = 10_000
    else:
        opts["num_seqs"] = 1_000
        opts["fixed_random_seed"] = sum(map(ord, name))
    return opts


train = _get_dataset_opts("train")
dev = _get_dataset_opts("dev")

targets_ext_dim = targets_dim + 1  # for BOS/EOS
targets_eos_idx = targets_ext_dim.dimension - 1
encoder_in_dim = Dim(80, name="enc_in_feature")


def get_model(**_kwargs):
    """get model, RETURNN config callback"""
    return Model(
        in_dim=data_feature_dim,
        encoder_in_dim=encoder_in_dim,
        num_enc_layers=12,
        enc_model_dim=Dim(name="enc", dimension=512, kind=Dim.Types.Feature),
        enc_ff_dim=Dim(name="enc-ff", dimension=2048, kind=Dim.Types.Feature),
        enc_att_num_heads=8,
        enc_conformer_layer_opts=dict(
            conv_norm_opts=dict(use_mask=True),
            self_att_opts=dict(
                # Shawn et al. 2018 style, old RETURNN way.
                with_bias=False,
                with_linear_pos=False,
                with_pos_bias=False,
                learnable_pos_emb=True,
                separate_pos_emb_per_head=False,
            ),
            ff_activation=lambda x: rf.relu(x) ** 2.0,
        ),
        target_dim=targets_ext_dim,
        bos_idx=targets_eos_idx,
        eos_idx=targets_eos_idx,
    )


class Model(rf.Module):
    """Model definition"""

    # noinspection PyShadowingNames
    def __init__(
        self,
        in_dim: Dim,
        encoder_in_dim: Dim,
        *,
        num_enc_layers: int = 12,
        target_dim: Dim,
        eos_idx: int,
        bos_idx: int,
        enc_model_dim: Dim = Dim(name="enc", dimension=512),
        enc_ff_dim: Dim = Dim(name="enc-ff", dimension=2048),
        enc_att_num_heads: int = 4,
        enc_conformer_layer_opts: Optional[Dict[str, Any]] = None,
        enc_key_total_dim: Dim = Dim(name="enc_key_total_dim", dimension=1024),
        att_num_heads: Dim = Dim(name="att_num_heads", dimension=1),
        att_dropout: float = 0.1,
        enc_dropout: float = 0.1,
        enc_att_dropout: float = 0.1,
        l2: float = 0.0001,
    ):
        super(Model, self).__init__()
        self.encoder_in_dim = encoder_in_dim
        self.embedding = rf.Embedding(in_dim, encoder_in_dim)
        self.encoder = ConformerEncoder(
            encoder_in_dim,
            enc_model_dim,
            ff_dim=enc_ff_dim,
            input_layer=ConformerConvSubsample(
                encoder_in_dim,
                out_dims=[Dim(32, name="conv1"), Dim(64, name="conv2"), Dim(64, name="conv3")],
                filter_sizes=[(3, 3), (3, 3), (3, 3)],
                pool_sizes=[(1, 2)],
                strides=[(1, 1), (3, 1), (2, 1)],
            ),
            encoder_layer_opts=enc_conformer_layer_opts,
            num_layers=num_enc_layers,
            num_heads=enc_att_num_heads,
            dropout=enc_dropout,
            att_dropout=enc_att_dropout,
        )

        self.target_dim = target_dim
        self.eos_idx = eos_idx
        self.bos_idx = bos_idx  # for non-blank labels; for with-blank labels, we use bos_idx=blank_idx

        self.enc_key_total_dim = enc_key_total_dim
        self.enc_key_per_head_dim = enc_key_total_dim.div_left(att_num_heads)
        self.att_num_heads = att_num_heads
        self.att_dropout = att_dropout
        self.dropout_broadcast = rf.dropout_broadcast_default()

        # https://github.com/rwth-i6/returnn-experiments/blob/master/2020-rnn-transducer/configs/base2.conv2l.specaug4a.ctc.devtrain.config

        self.enc_ctx = rf.Linear(self.encoder.out_dim, enc_key_total_dim)
        self.enc_ctx_dropout = 0.2
        self.enc_win_dim = Dim(name="enc_win_dim", dimension=5)

        self.inv_fertility = rf.Linear(self.encoder.out_dim, att_num_heads, with_bias=False)

        self.target_embed = rf.Embedding(target_dim, Dim(name="target_embed", dimension=640))

        self.s = rf.ZoneoutLSTM(
            self.target_embed.out_dim + att_num_heads * self.encoder.out_dim,
            Dim(name="lstm", dimension=1024),
            zoneout_factor_cell=0.15,
            zoneout_factor_output=0.05,
            use_zoneout_output=False,  # like RETURNN/TF ZoneoutLSTM old default
            # parts_order="icfo",  # like RETURNN/TF ZoneoutLSTM
            # parts_order="ifco",
            parts_order="jifo",  # NativeLSTM (the code above converts it...)
            forget_bias=0.0,  # the code above already adds it during conversion
        )

        self.weight_feedback = rf.Linear(att_num_heads, enc_key_total_dim, with_bias=False)
        self.s_transformed = rf.Linear(self.s.out_dim, enc_key_total_dim, with_bias=False)
        self.energy = rf.Linear(enc_key_total_dim, att_num_heads, with_bias=False)
        self.readout_in = rf.Linear(
            self.s.out_dim + self.target_embed.out_dim + att_num_heads * self.encoder.out_dim,
            Dim(name="readout", dimension=1024),
        )
        self.output_prob = rf.Linear(self.readout_in.out_dim // 2, target_dim)

        for p in self.parameters():
            p.weight_decay = l2

    def encode(
        self,
        source: Tensor,
        *,
        in_spatial_dim: Dim,
        collected_outputs: Optional[Dict[str, Tensor]] = None,
    ) -> Tuple[Dict[str, Tensor], Dim]:
        """encode, and extend the encoder output for things we need in the decoder"""
        source = self.embedding(source)
        # SpecAugment
        source = rf.audio.specaugment(source, spatial_dim=in_spatial_dim, feature_dim=self.encoder_in_dim)
        # Encoder including convolutional frontend
        enc, enc_spatial_dim = self.encoder(source, in_spatial_dim=in_spatial_dim, collected_outputs=collected_outputs)
        enc_ctx = self.enc_ctx(enc)
        inv_fertility = rf.sigmoid(self.inv_fertility(enc))
        return dict(enc=enc, enc_ctx=enc_ctx, inv_fertility=inv_fertility), enc_spatial_dim

    def decoder_default_initial_state(self, *, batch_dims: Sequence[Dim], enc_spatial_dim: Dim) -> rf.State:
        """Default initial state"""
        state = rf.State(
            s=self.s.default_initial_state(batch_dims=batch_dims),
            att=rf.zeros(list(batch_dims) + [self.att_num_heads * self.encoder.out_dim]),
            accum_att_weights=rf.zeros(
                list(batch_dims) + [enc_spatial_dim, self.att_num_heads], feature_dim=self.att_num_heads
            ),
        )
        state.att.feature_dim_axis = len(state.att.dims) - 1
        return state

    def loop_step_output_templates(self, batch_dims: List[Dim]) -> Dict[str, Tensor]:
        """loop step out"""
        return {
            "s": Tensor(
                "s", dims=batch_dims + [self.s.out_dim], dtype=rf.get_default_float_dtype(), feature_dim_axis=-1
            ),
            "att": Tensor(
                "att",
                dims=batch_dims + [self.att_num_heads * self.encoder.out_dim],
                dtype=rf.get_default_float_dtype(),
                feature_dim_axis=-1,
            ),
        }

    def loop_step(
        self,
        *,
        enc: rf.Tensor,
        enc_ctx: rf.Tensor,
        inv_fertility: rf.Tensor,
        enc_spatial_dim: Dim,
        input_embed: rf.Tensor,
        state: Optional[rf.State] = None,
    ) -> Tuple[Dict[str, rf.Tensor], rf.State]:
        """step of the inner loop"""
        if state is None:
            batch_dims = enc.remaining_dims(
                remove=(enc.feature_dim, enc_spatial_dim) if enc_spatial_dim != single_step_dim else (enc.feature_dim,)
            )
            state = self.decoder_default_initial_state(batch_dims=batch_dims, enc_spatial_dim=enc_spatial_dim)
        state_ = rf.State()

        prev_att = state.att

        s, state_.s = self.s(rf.concat_features(input_embed, prev_att), state=state.s, spatial_dim=single_step_dim)

        weight_feedback = self.weight_feedback(state.accum_att_weights)
        s_transformed = self.s_transformed(s)
        energy_in = enc_ctx + weight_feedback + s_transformed
        energy = self.energy(rf.tanh(energy_in))
        att_weights = rf.softmax(energy, axis=enc_spatial_dim)
        state_.accum_att_weights = state.accum_att_weights + att_weights * inv_fertility * 0.5
        att0 = rf.dot(att_weights, enc, reduce=enc_spatial_dim, use_mask=False)
        att0.feature_dim = self.encoder.out_dim
        att, _ = rf.merge_dims(att0, dims=(self.att_num_heads, self.encoder.out_dim))
        state_.att = att

        return {"s": s, "att": att}, state_

    def decode_logits(self, *, s: Tensor, input_embed: Tensor, att: Tensor) -> Tensor:
        """logits for the decoder"""
        readout_in = self.readout_in(rf.concat_features(s, input_embed, att))
        readout = rf.reduce_out(readout_in, mode="max", num_pieces=2, out_dim=self.output_prob.in_dim)
        readout = rf.dropout(readout, drop_prob=0.3, axis=self.dropout_broadcast and readout.feature_dim)
        logits = self.output_prob(readout)
        return logits


# noinspection PyShadowingNames
def train_step(*, model: Model, extern_data: TensorDict, **_kwargs):
    """Function is run within RETURNN."""
    data = extern_data[extern_data_inputs_name]
    targets = extern_data[extern_data_targets_name]
    enc_args, enc_spatial_dim = model.encode(data, in_spatial_dim=data_spatial_dim)

    batch_dims = data.remaining_dims(data_spatial_dim)
    targets = targets.copy()
    targets.sparse_dim = targets_ext_dim
    targets = rf.shift_right(targets, axis=targets_spatial_dim, pad_value=targets_eos_idx)
    input_embeddings = model.target_embed(targets)

    def _body(input_embed: Tensor, state: rf.State):
        new_state = rf.State()
        loop_out_, new_state.decoder = model.loop_step(
            **enc_args,
            enc_spatial_dim=enc_spatial_dim,
            input_embed=input_embed,
            state=state.decoder,
        )
        return loop_out_, new_state

    loop_out, _, _ = rf.scan(
        spatial_dim=targets_spatial_dim,
        xs=input_embeddings,
        ys=model.loop_step_output_templates(batch_dims=batch_dims),
        initial=rf.State(
            decoder=model.decoder_default_initial_state(batch_dims=batch_dims, enc_spatial_dim=enc_spatial_dim),
        ),
        body=_body,
    )

    logits = model.decode_logits(input_embed=input_embeddings, **loop_out)
    logits_packed, pack_dim = rf.pack_padded(logits, dims=batch_dims + [targets_spatial_dim], enforce_sorted=False)
    targets_packed, _ = rf.pack_padded(
        targets, dims=batch_dims + [targets_spatial_dim], enforce_sorted=False, out_dim=pack_dim
    )

    log_prob = rf.log_softmax(logits_packed, axis=model.target_dim)
    log_prob = rf.label_smoothed_log_prob_gradient(log_prob, 0.1, axis=model.target_dim)
    loss = rf.cross_entropy(
        target=targets_packed, estimated=log_prob, estimated_type="log-probs", axis=model.target_dim
    )
    loss.mark_as_loss("ce")

    best = rf.reduce_argmax(logits_packed, axis=model.target_dim)
    frame_error = best != targets_packed
    frame_error.mark_as_loss(name="fer", as_error=True)


# noinspection PyShadowingNames
def forward_step(*, model: Model, extern_data: TensorDict, **_kwargs) -> Tuple[Tensor, Tensor, Dim, Dim]:
    """
    Function is run within RETURNN.

    Earlier we used the generic beam_search function,
    but now we just directly perform the search here,
    as this is overall simpler and shorter.

    :return:
        recog results including beam {batch, beam, out_spatial},
        log probs {batch, beam},
        out_spatial_dim,
        final beam_dim
    """
    data = extern_data[extern_data_inputs_name]
    batch_dims = data.remaining_dims((data_spatial_dim, data.feature_dim))
    enc_args, enc_spatial_dim = model.encode(data, in_spatial_dim=data_spatial_dim)
    beam_size = 12
    length_normalization_exponent = 1.0
    max_seq_len = enc_spatial_dim.get_size_tensor()
    print("** max seq len:", max_seq_len.raw_tensor)

    # Eager-mode implementation of beam search.
    # Initial state.
    beam_dim = Dim(1, name="initial-beam")
    batch_dims_ = [beam_dim] + batch_dims
    decoder_state = model.decoder_default_initial_state(batch_dims=batch_dims_, enc_spatial_dim=enc_spatial_dim)
    target = rf.constant(model.bos_idx, dims=batch_dims_, sparse_dim=model.target_dim)
    ended = rf.constant(False, dims=batch_dims_)
    out_seq_len = rf.constant(0, dims=batch_dims_)
    seq_log_prob = rf.constant(0.0, dims=batch_dims_)

    i = 0
    seq_targets = []
    seq_backrefs = []
    while True:
        input_embed = model.target_embed(target)
        step_out, decoder_state = model.loop_step(
            **enc_args,
            enc_spatial_dim=enc_spatial_dim,
            input_embed=input_embed,
            state=decoder_state,
        )
        logits = model.decode_logits(input_embed=input_embed, **step_out)
        label_log_prob = rf.log_softmax(logits, axis=model.target_dim)
        # Filter out finished beams
        label_log_prob = rf.where(
            ended,
            rf.sparse_to_dense(model.eos_idx, axis=model.target_dim, label_value=0.0, other_value=-1.0e30),
            label_log_prob,
        )
        seq_log_prob = seq_log_prob + label_log_prob  # Batch, InBeam, Vocab
        seq_log_prob, (backrefs, target), beam_dim = rf.top_k(
            seq_log_prob, k_dim=Dim(beam_size, name=f"dec-step{i}-beam"), axis=[beam_dim, model.target_dim]
        )  # seq_log_prob, backrefs, target: Batch, Beam
        seq_targets.append(target)
        seq_backrefs.append(backrefs)
        decoder_state = tree.map_structure(lambda s: rf.gather(s, indices=backrefs), decoder_state)
        ended = rf.gather(ended, indices=backrefs)
        out_seq_len = rf.gather(out_seq_len, indices=backrefs)
        i += 1

        ended = rf.logical_or(ended, target == model.eos_idx)
        ended = rf.logical_or(ended, rf.copy_to_device(i >= max_seq_len))
        if bool(rf.reduce_all(ended, axis=ended.dims).raw_tensor):
            break
        out_seq_len = out_seq_len + rf.where(ended, 0, 1)

        if i > 1 and length_normalization_exponent != 0:
            # Length-normalized scores, so we evaluate score_t/len.
            # If seq ended, score_i/i == score_{i-1}/(i-1), thus score_i = score_{i-1}*(i/(i-1))
            # Because we count with EOS symbol, shifted by one.
            seq_log_prob *= rf.where(
                ended,
                (i / (i - 1)) ** length_normalization_exponent,
                1.0,
            )

    if i > 0 and length_normalization_exponent != 0:
        seq_log_prob *= (1 / i) ** length_normalization_exponent

    # Backtrack via backrefs, resolve beams.
    seq_targets_ = []
    indices = rf.range_over_dim(beam_dim)  # FinalBeam -> FinalBeam
    for backrefs, target in zip(seq_backrefs[::-1], seq_targets[::-1]):
        # indices: FinalBeam -> Beam
        # backrefs: Beam -> PrevBeam
        seq_targets_.insert(0, rf.gather(target, indices=indices))
        indices = rf.gather(backrefs, indices=indices)  # FinalBeam -> PrevBeam

    seq_targets__ = TensorArray(seq_targets_[0])
    for target in seq_targets_:
        seq_targets__ = seq_targets__.push_back(target)
    out_spatial_dim = Dim(out_seq_len, name="out-spatial")
    seq_targets = seq_targets__.stack(axis=out_spatial_dim)

    return seq_targets, seq_log_prob, out_spatial_dim, beam_dim


# See: https://github.com/rwth-i6/returnn/issues/1402
_interesting_commits = [
    "a776227",  # -- some baseline before those first optimizations
    "f09222e",
    "fa9818c",
    "01d0653",
    "dc14a2c",
    "361e238",
    "49b69ed",
    "07078b9",
    "2e104f5",
]


def main():
    """main"""
    # pass ++num_epochs 1 ++device cpu or so...
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--bench-action", choices=("run", "multi-run", "profile"), default="run")
    args, remaining_args = arg_parser.parse_known_args()

    try:
        # https://youtrack.jetbrains.com/issue/PY-63226/PyCharm-debugger-hangs-when-process-is-forking
        multiprocessing.set_start_method("spawn")
    except Exception as exc:
        print("multiprocessing.set_start_method 'spawn' exception:", exc)
        print("Ignoring this...")

    if args.bench_action == "run":
        __main__.main(sys.argv[:1] + [_my_file] + remaining_args)
    elif args.bench_action == "profile":
        _custom_loop(remaining_args)
    elif args.bench_action == "multi-run":
        with tempfile.TemporaryDirectory(prefix="returnn-tmp-checkout-") as returnn_tmp_dir:
            _subproc_check_call("git", "clone", "--shared", _returnn_root_dir, returnn_tmp_dir)
            os.chdir(returnn_tmp_dir)
            _subproc_check_call("git", "config", "--local", "advice.detachedHead", "false")
            os.environ["PYTHONUNBUFFERED"] = "1"
            for commit in _interesting_commits:
                _subproc_check_call("git", "checkout", commit)
                _subproc_check_call_filter_returnn_out(sys.executable, "rnn.py", _my_file, *remaining_args)
    else:
        raise ValueError(f"invalid --bench-action {args.bench_action!r}")


def _custom_loop(argv):
    from returnn.log import log
    from returnn.util.basic import hms
    from returnn.datasets import init_dataset
    from returnn.torch.data import pipeline as data_pipeline
    from returnn.torch.data import returnn_dataset_wrapper
    from returnn.torch.data import extern_data as extern_data_util
    from returnn.torch.engine import get_device_from_config_opt
    from returnn.torch.frontend.bridge import rf_module_to_pt_module

    import torch
    from torch.utils.data import DataLoader

    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--device", default=None)
    arg_parser.add_argument("--tb-dir", default="tb-log")
    args = arg_parser.parse_args(argv)

    rf.select_backend_torch()

    extern_data_template = extern_data_util.extern_data_template_from_config_opts(extern_data)

    device_with_reason = get_device_from_config_opt(args.device)
    device = device_with_reason.result
    print("Using device:", device, f"({device_with_reason.reason})", file=log.v2)

    model = get_model()
    pt_model = rf_module_to_pt_module(model)
    pt_model.to(device)
    pt_model.train()

    optimizer = torch.optim.Adam(pt_model.parameters(), lr=config["learning_rate"])

    dataset = init_dataset(train)
    wrapped_dataset = returnn_dataset_wrapper.ReturnnDatasetIterDataPipe(dataset)
    batch_size = config["batch_size"]
    max_seqs = config["max_seqs"]
    batches_dataset = data_pipeline.BatchingIterDataPipe(wrapped_dataset, batch_size=batch_size, max_seqs=max_seqs)
    data_loader = DataLoader(batches_dataset, batch_size=None, collate_fn=data_pipeline.collate_batch)
    data_iter = iter(data_loader)

    # noinspection PyUnresolvedReferences,PyProtectedMember
    with torch.profiler.profile(
        schedule=torch.profiler.schedule(wait=1, warmup=4, active=3, repeat=2),
        on_trace_ready=torch.profiler.tensorboard_trace_handler(args.tb_dir),
        record_shapes=True,
        profile_memory=True,
        with_stack=True,
        # https://github.com/pytorch/pytorch/issues/100253
        experimental_config=torch._C._profiler._ExperimentalConfig(verbose=True),
    ) as prof:
        step_idx = 0
        epoch_start_time = time.time()
        elapsed_computation_time = 0

        while True:
            with torch.no_grad():
                extern_data_raw = next(data_iter, None)
            if extern_data_raw is None:
                break

            step_begin_time = time.time()

            optimizer.zero_grad()

            extern_data_ = extern_data_util.raw_dict_to_extern_data(
                extern_data_raw, extern_data_template=extern_data_template, device=device
            )
            rf.init_train_step_run_ctx(train_flag=True, step=step_idx, epoch=1)
            with rf.set_default_device_ctx(device):
                train_step(model=model, extern_data=extern_data_)

            train_ctx = rf.get_run_ctx()
            total_loss = train_ctx.total_loss()
            total_loss.raw_tensor.backward()

            optimizer.step()

            print("step %i, loss %f" % (step_idx, total_loss.raw_tensor.detach().cpu()), file=log.v3)

            elapsed_computation_time += time.time() - step_begin_time
            step_idx += 1
            prof.step()

        elapsed = time.time() - epoch_start_time
        elapsed_computation_percentage = elapsed_computation_time / elapsed
        print(
            "Trained %i steps, %s elapsed (%.1f%% computing time)"
            % (step_idx, hms(elapsed), (elapsed_computation_percentage * 100.0)),
            file=log.v3,
        )


def _subproc_check_call(*args):
    print("$", *args)
    subprocess.check_call(args)


def _subproc_check_call_filter_returnn_out(*args):
    print("$", *args)
    line_count = 0
    need_newline = False
    with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as p:
        while True:
            line = p.stdout.readline()
            if not line:
                break
            if line_count < 5:
                sys.stdout.buffer.write(line)
            # "Trained %i steps, %s elapsed (%.1f%% computing time)"
            # "elapsed: ..."
            elif b"elapsed" in line:
                if need_newline:
                    sys.stdout.buffer.write(b"\n")
                sys.stdout.buffer.write(line)
            else:
                sys.stdout.buffer.write(b".")
                need_newline = True
            sys.stdout.buffer.flush()
            line_count += 1
    if need_newline:
        sys.stdout.buffer.write(b"\n")
        sys.stdout.buffer.flush()
    if p.returncode != 0:
        raise subprocess.CalledProcessError(p.returncode, args)


if __name__ == "__main__":
    main()

else:
    globals().update(config)



"""
E-Branchformer (https://arxiv.org/pdf/2210.00077)

Example usage:

.. code-block:: python

    import returnn.frontend as rf
    from returnn.frontend.encoder.conformer import ConformerEncoder
    from returnn.frontend.encoder.e_branchformer import EBranchformerLayer

    model = ConformerEncoder(
        out_dim=...,  # model dim, output_size in ESPnet
        num_layers=...,  # num_blocks in ESPnet
        encoder_layer=rf.build_dict(
            EBranchformerLayer,
            ff_dim=...,  # linear_units in ESPnet
            num_heads=...,  # attention_heads in ESPnet
            cgmlp_ff_dim=...,  # half of cgmlp_linear_units in ESPnet
            cgmlp_conv_kernel=...,  # cgmlp_conv_kernel in ESPnet
            merge_conv_kernel=...,  # merge_conv_kernel in ESPnet
        ),
    )

"""

from __future__ import annotations
from typing import Union, Any, Callable, Dict
import copy as _copy
from returnn.util.basic import NotSpecified
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from ..decoder.transformer import make_norm
from .conformer import make_ff


__all__ = ["EBranchformerLayer", "FeedForwardConvGated", "Merge"]


class EBranchformerLayer(rf.Module):
    """
    E-Branchformer layer, e.g. to be used in the :class:`returnn.frontend.encoder.conformer.ConformerEncoder`.

    See the module docstring :mod:`returnn.frontend.encoder.e_branchformer` for an example.
    """

    def __init__(
        self,
        out_dim: Dim = Dim(512, name="conformer-enc-default-out-dim"),
        *,
        ff: Union[type, Dict[str, Any], rf.Module] = NotSpecified,
        ff_dim: Union[Dim, int] = NotSpecified,
        ff_activation: Union[Callable[[Tensor], Tensor], Dict[str, Any], rf.Module] = NotSpecified,
        dropout: float = 0.1,
        num_heads: int = 4,
        self_att: Union[rf.RelPosSelfAttention, rf.Module, type, Dict[str, Any], Any] = NotSpecified,
        att_dropout: float = 0.1,
        cgmlp: Union[type, Dict[str, Any]] = NotSpecified,
        cgmlp_ff_dim: Union[Dim, int] = NotSpecified,
        cgmlp_conv_kernel: int = 31,
        merge_conv_kernel: int = 3,
        norm: Union[type, Dict[str, Any], rf.Module, Callable] = rf.LayerNorm,
    ):
        super().__init__()

        self.dropout = dropout
        self.dropout_broadcast = rf.dropout_broadcast_default()
        self.out_dim = out_dim

        self.ffn1 = make_ff(ff=ff, out_dim=out_dim, ff_dim=ff_dim, dropout=dropout, ff_activation=ff_activation)
        self.ffn1_layer_norm = make_norm(norm, out_dim)

        self.ffn2 = make_ff(ff=ff, out_dim=out_dim, ff_dim=ff_dim, dropout=dropout, ff_activation=ff_activation)
        self.ffn2_layer_norm = make_norm(norm, out_dim)

        if self_att is NotSpecified or isinstance(self_att, (dict, type)):
            self_att_opts_ = dict(
                in_dim=out_dim,
                proj_dim=out_dim,
                key_dim_total=out_dim,
                value_dim_total=out_dim,
                num_heads=num_heads,
                att_dropout=att_dropout,
            )
            self_att_opts_ = {k: v for (k, v) in self_att_opts_.items() if v is not NotSpecified}
            if self_att is NotSpecified:
                self.self_att = rf.RelPosSelfAttention(**self_att_opts_)
            elif isinstance(self_att, type):
                self.self_att = self_att(**self_att_opts_)
            elif isinstance(self_att, dict):
                self_att_opts_ = {k: v for (k, v) in self_att_opts_.items() if k not in self_att}
                self.self_att = rf.build_from_dict(self_att, **self_att_opts_)
            else:
                raise TypeError(f"{self}: invalid type: self_att {self_att!r}")
        elif isinstance(self_att, rf.Module):
            self.self_att = _copy.deepcopy(self_att)
        else:
            raise TypeError(f"{self}: invalid self_att {self_att!r} type {type(self_att).__name__}")
        self.self_att_layer_norm = make_norm(norm, out_dim)

        if cgmlp is NotSpecified:
            cgmlp = FeedForwardConvGated
        cgmlp_opts = {"ff_dim": cgmlp_ff_dim, "kernel_size": cgmlp_conv_kernel, "dropout": dropout}
        cgmlp_opts = {k: v for (k, v) in cgmlp_opts.items() if v is not NotSpecified}
        if isinstance(cgmlp, type):
            cgmlp = cgmlp(out_dim, **cgmlp_opts)
        elif isinstance(cgmlp, dict):
            cgmlp_opts = {k: v for (k, v) in cgmlp_opts.items() if k not in cgmlp}
            cgmlp = rf.build_from_dict(cgmlp, out_dim, **cgmlp_opts)
        elif isinstance(cgmlp, rf.Module):
            cgmlp = _copy.deepcopy(cgmlp)
        if not isinstance(cgmlp, rf.Module):
            raise TypeError(f"unexpected cgmlp {cgmlp!r} type {type(cgmlp).__name__}")
        self.cgmlp: Union[FeedForwardConvGated, rf.Module] = cgmlp
        self.cgmlp_layer_norm = make_norm(norm, out_dim)

        self.merge = Merge(in_dim1=out_dim, in_dim2=out_dim, out_dim=out_dim, merge_conv_kernel=merge_conv_kernel)

        self.final_layer_norm = make_norm(norm, out_dim)

    def __call__(self, inp: Tensor, *, spatial_dim: Dim) -> Tensor:
        """forward"""
        # FFN
        x_ffn1_ln = self.ffn1_layer_norm(inp)
        x_ffn1 = self.ffn1(x_ffn1_ln)
        x_ffn1_out = 0.5 * rf.dropout(x_ffn1, self.dropout, axis=self.dropout_broadcast and self.out_dim) + inp

        # Branch 1: MHSA
        x_mhsa_ln = self.self_att_layer_norm(x_ffn1_out)
        x_mhsa = self.self_att(x_mhsa_ln, axis=spatial_dim)
        x_mhsa = rf.dropout(x_mhsa, self.dropout, axis=self.dropout_broadcast and self.out_dim)

        # Branch 2: cgMLP
        x_cgmlp_ln = self.cgmlp_layer_norm(x_ffn1_out)
        x_cgmlp = self.cgmlp(x_cgmlp_ln, spatial_dim=spatial_dim)
        x_cgmlp = rf.dropout(x_cgmlp, self.dropout, axis=self.dropout_broadcast and self.out_dim)

        # Merge branches
        x_merge = self.merge(x_mhsa, x_cgmlp, spatial_dim=spatial_dim)
        x_merge = rf.dropout(x_merge, self.dropout, axis=self.dropout_broadcast and self.out_dim)
        x_merge_out = x_merge + x_ffn1_out

        # FFN
        x_ffn2_ln = self.ffn2_layer_norm(x_merge_out)
        x_ffn2 = self.ffn2(x_ffn2_ln)
        x_ffn2_out = 0.5 * rf.dropout(x_ffn2, self.dropout, axis=self.dropout_broadcast and self.out_dim) + x_merge_out

        # last LN layer
        return self.final_layer_norm(x_ffn2_out)


class FeedForwardConvGated(rf.Module):
    """
    Convolutional Gating MLP (cgMLP) as introduced in https://openreview.net/forum?id=RA-zVvZLYIy
    and then used by the E-Branchformer model (https://arxiv.org/pdf/2210.00077).
    It uses the Convolutional Spatial Gating Unit (CSGU).
    This is the local extractor branch in the E-Branchformer model.

    Related is the :class:`returnn.frontend.decoder.transformer.FeedForwardGated` module.
    """

    def __init__(
        self,
        out_dim: Dim,
        *,
        ff_dim: Union[Dim, int] = NotSpecified,
        kernel_size: int = 31,
        dropout: float = 0.1,
        activation: Union[Callable[[Tensor], Tensor], Dict[str, Any], rf.Module] = rf.gelu,
        gate_activation: Union[Callable[[Tensor], Tensor], Dict[str, Any], rf.Module] = rf.identity,
        with_bias: bool = True,
        norm: Union[type, Dict[str, Any], rf.Module, Callable] = rf.LayerNorm,
    ):
        """
        :param out_dim: the encoder (e.g. E-Branchformer) model dim. (usually 256 or 512)
        :param ff_dim: intermediate dimension.
            This is like cgmlp_linear_units/2 in ESPnet.
            Note the 1/2 factor, which is because in ESPnet, you specify the total dimension,
            before it is split for the gating,
            while here, you specify the dimension for the gating part.
            Common settings are 2048/2 or 3072/2.
            In the paper, they mention a factor of 3 of the model dimension (factor 6 for ESPnet setting).
        :param kernel_size: for the depthwise convolution (usually 31)
        :param dropout:
        :param activation: activation function after the first linear layer, for both parts.
            default as in the paper: gelu.
            Note, in :class:`returnn.frontend.decoder.transformer.FeedForwardGated`,
            the ``activation`` arg is like ``gate_activation`` here.
        :param gate_activation: activation function for the gate part, before the gating (mult) is applied.
            default as in the paper: identity.
            Note, in :class:`returnn.frontend.decoder.transformer.FeedForwardGated`,
            the ``activation`` arg is like ``gate_activation`` here.
        :param with_bias: whether to use bias in the linear layers and conv layer. default as in the paper: True.
        :param norm: normalization layer. default as in the paper: LayerNorm.
        """
        super().__init__()

        if ff_dim is NotSpecified:
            ff_dim = out_dim * 3  # somewhat arbitrary. with 512, this is 3072/2.
        if isinstance(ff_dim, int):
            ff_dim = Dim(ff_dim, name="e-branchformer-ff-dim")
        if not isinstance(ff_dim, Dim):
            raise TypeError(f"E-Branchformer FeedForwardConvGated: unexpected ff_dim {ff_dim!r} type {type(ff_dim)}")

        self.out_dim = out_dim
        self.ff_dim = ff_dim
        self.dropout = dropout
        self.dropout_broadcast = rf.dropout_broadcast_default()
        if activation is NotSpecified:
            activation = rf.gelu
        self.activation = _make_activation(activation)
        if gate_activation is NotSpecified:
            gate_activation = rf.identity
        self.gate_activation = _make_activation(gate_activation)

        self.linear_ff = rf.Linear(out_dim, 2 * ff_dim, with_bias=with_bias)  # x2 to split for gating
        self.norm = make_norm(norm, ff_dim)
        self.conv = rf.Conv1d(  # depthwise convolution
            ff_dim, ff_dim, filter_size=kernel_size, groups=ff_dim.dimension, padding="same", with_bias=with_bias
        )
        self.linear_out = rf.Linear(ff_dim, out_dim, with_bias=with_bias)

    def __call__(self, x: Tensor, *, spatial_dim: Dim) -> Tensor:
        """forward"""
        x = self.linear_ff(x)
        x = self.activation(x)

        # Now the convolutional spatial gating part (CSGU).
        # Note: Different order than in ESPnet ConvolutionalSpatialGatingUnit,
        # but consistent to our own FeedForwardGated.
        x_g, x_r = rf.split(x, axis=self.linear_ff.out_dim, out_dims=[self.ff_dim] * 2)
        x_g = self.norm(x_g)
        x_g, _ = self.conv(x_g, in_spatial_dim=spatial_dim)
        x_g = self.gate_activation(x_g)
        x = x_g * x_r

        x = rf.dropout(x, self.dropout, axis=self.dropout_broadcast and self.linear_out.in_dim)
        x = self.linear_out(x)
        return x


class Merge(rf.Module):
    """
    The merge module from the E-Branchformer model.
    """

    def __init__(self, *, in_dim1: Dim, in_dim2: Dim, out_dim: Dim, merge_conv_kernel: int = 3):
        super().__init__()
        self.in_dim1 = in_dim1
        self.in_dim2 = in_dim2
        self.concat_in_dim = in_dim1 + in_dim2
        self.out_dim = out_dim
        self.depthwise_conv_fusion = rf.Conv1d(
            self.concat_in_dim,
            self.concat_in_dim,
            groups=self.concat_in_dim.dimension,
            filter_size=merge_conv_kernel,
            padding="same",
        )
        self.merge_proj = rf.Linear(self.concat_in_dim, out_dim)

    def __call__(self, x1: Tensor, x2: Tensor, *, spatial_dim: Dim) -> Tensor:
        """forward"""
        x_concat, _ = rf.concat((x1, self.in_dim1), (x2, self.in_dim2), out_dim=self.concat_in_dim)
        x_conv, _ = self.depthwise_conv_fusion(x_concat, in_spatial_dim=spatial_dim)
        y = self.merge_proj(x_concat + x_conv)
        return y


def _make_activation(
    activation: Union[Callable[[Tensor], Tensor], Dict[str, Any], rf.Module]
) -> Union[Callable[[Tensor], Tensor], rf.Module]:
    if isinstance(activation, dict):
        activation = rf.build_from_dict(activation)
    elif isinstance(activation, rf.Module):
        activation = _copy.deepcopy(activation)
    elif isinstance(activation, type):
        activation = activation()
    if not callable(activation):
        raise TypeError(f"unexpected activation type {activation!r}")
    return activation



"""
Conformer model, variant of Transformer with additional convolution, introduced for speech recognition.
Ref: https://arxiv.org/abs/2005.08100

About details of the specific implementation and other implementations, see:
https://github.com/rwth-i6/returnn_common/issues/233

V2: Split frontend and main encoder.
"""

from __future__ import annotations
from typing import Optional, Union, Any, Tuple, Dict, Callable
import copy as _copy
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from returnn.util.basic import NotSpecified
from .base import ISeqDownsamplingEncoder, ISeqFramewiseEncoder
from .conformer import ConformerConvSubsample, ConformerEncoderLayer


class ConformerFrontend(ISeqDownsamplingEncoder):
    """
    This is just the combination of:
    - input_layer (ConformerConvSubsample)
    - input_projection (Linear without bias)
    - input_embedding_scale
    - input_dropout

    This is intended to be used together with :class:`ConformerEncoderV2`.
    """

    def __init__(
        self,
        in_dim: Dim,
        out_dim: Dim = Dim(512, name="conformer-enc-default-out-dim"),
        *,
        input_layer: Optional[Union[ConformerConvSubsample, ISeqDownsamplingEncoder, rf.Module, Any]],
        input_embedding_scale: float = 1.0,
        input_dropout: float = 0.1,
    ):
        """
        :param in_dim: input features (e.g. MFCC)
        :param out_dim: the output feature dimension
        :param input_layer: input/frontend/prenet with potential subsampling.
            (x, in_spatial_dim) -> (y, out_spatial_dim)
        :param input_embedding_scale: applied after input_layer. 1.0 by default for historic reasons.
            In std Transformer, also ESPnet E-Branchformer and Conformer, this is sqrt(out_dim).
        :param input_dropout: applied after input_projection(input_layer(x))
        """
        super().__init__()

        self.in_dim = in_dim
        self.out_dim = out_dim
        self.dropout_broadcast = rf.dropout_broadcast_default()

        # TODO once we figured out good defaults, we would create ConformerConvSubsample here when not given
        if callable(input_layer) or input_layer is None:
            pass  # leave it as is
        elif isinstance(input_layer, dict):
            input_layer = rf.build_from_dict(input_layer, in_dim)
            input_layer: ConformerConvSubsample  # maybe not true, but assume for some attribs
        else:
            raise TypeError(f"unexpected input_layer {input_layer!r}")
        self.input_layer = input_layer
        self.input_projection = (
            rf.Linear(self.input_layer.out_dim if self.input_layer else self.in_dim, self.out_dim, with_bias=False)
            if input_layer
            else None
        )
        self.input_embedding_scale = input_embedding_scale
        self.input_dropout = input_dropout

    def __call__(self, source: Tensor, *, in_spatial_dim: Dim) -> Tuple[Tensor, Dim]:
        """forward"""
        if self.input_layer:
            x_subsample, out_spatial_dim = self.input_layer(source, in_spatial_dim=in_spatial_dim)
        else:
            x_subsample, out_spatial_dim = source, in_spatial_dim
        x = self.input_projection(x_subsample) if self.input_projection else x_subsample
        if self.input_embedding_scale != 1.0:
            x = x * self.input_embedding_scale
        x = rf.dropout(x, self.input_dropout, axis=self.dropout_broadcast and self.out_dim)
        return x, out_spatial_dim


class ConformerEncoderV2(ISeqFramewiseEncoder):
    """
    Conformer encoder, without frontend.

    V2: Without the input_layer / frontend module, i.e. just the conformer layers.
    Use ConformerFrontend for the frontend.
    To get the V1 case, add this in front:
    """

    def __init__(
        self,
        out_dim: Dim = Dim(512, name="conformer-enc-default-out-dim"),
        *,
        num_layers: int,
        ff_dim: Dim = NotSpecified,
        ff_activation: Union[Callable[[Tensor], Tensor], Dict[str, Any], rf.Module] = NotSpecified,
        dropout: float = 0.1,
        conv_kernel_size: int = NotSpecified,
        conv_norm: Union[rf.BatchNorm, type, Dict[str, Any], Any] = NotSpecified,
        num_heads: int = 4,
        att_dropout: float = 0.1,
        encoder_layer: Optional[Union[ConformerEncoderLayer, rf.Module, type, Dict[str, Any], Any]] = None,
        encoder_layer_opts: Optional[Dict[str, Any]] = None,
        sequential=rf.Sequential,
    ):
        """
        :param out_dim: the output feature dimension
        :param num_layers: the number of encoder layers
        :param ff_dim: the dimension of feed-forward layers. 2048 originally, or 4 times out_dim
        :param ff_activation: activation function for feed-forward network
        :param dropout: the dropout value for the FF block
        :param conv_kernel_size: the kernel size of depthwise convolution in the conv block
        :param conv_norm: used for the conv block. Batch norm originally
        :param num_heads: the number of attention heads
        :param att_dropout: attention dropout value
        :param encoder_layer: an instance of :class:`ConformerEncoderLayer` or similar
        :param encoder_layer_opts: options for the encoder layer
        :param sequential:
        """
        super().__init__()

        self.out_dim = out_dim
        self.dropout = dropout
        self.dropout_broadcast = rf.dropout_broadcast_default()

        if not encoder_layer or isinstance(encoder_layer, (dict, type)):
            encoder_layer_opts_ = dict(
                out_dim=out_dim,
                ff_dim=ff_dim,
                ff_activation=ff_activation,
                dropout=dropout,
                conv_kernel_size=conv_kernel_size,
                conv_norm=conv_norm,
                num_heads=num_heads,
                att_dropout=att_dropout,
            )
            encoder_layer_opts_ = {k: v for (k, v) in encoder_layer_opts_.items() if v is not NotSpecified}
            if encoder_layer_opts:
                encoder_layer_opts_.update(encoder_layer_opts)
            if not encoder_layer:
                encoder_layer = ConformerEncoderLayer(**encoder_layer_opts_)
            elif isinstance(encoder_layer, type):
                encoder_layer = encoder_layer(**encoder_layer_opts_)
            elif isinstance(encoder_layer, dict):
                # Note: Reuse all the encoder_layer_opts_.
                # If this does not make sense for the specific encoder_layer class here,
                # we would suggest to use a different ConformerEncoder class.
                # (The alternative, to not reuse encoder_layer_opts_ here,
                #  would probably be more confusing, as those options are all ignored then.
                #  It's also not clear what args to pass then and what not.)
                # (Maybe we should do a ConformerEncoderV2 if this is confusing here...)
                encoder_layer_opts_ = {k: v for (k, v) in encoder_layer_opts_.items() if k not in encoder_layer}
                encoder_layer = rf.build_from_dict(encoder_layer, **encoder_layer_opts_)
            else:
                raise TypeError(f"unexpected encoder_layer {encoder_layer!r}")
        else:
            if not callable(encoder_layer):
                raise TypeError(f"{self}: invalid non-callable encoder_layer {encoder_layer!r}")

        self.layers = sequential(_copy.deepcopy(encoder_layer) for _ in range(num_layers))

    def __call__(self, x: Tensor, *, spatial_dim: Dim, collected_outputs: Optional[Dict[str, Tensor]] = None) -> Tensor:
        """forward"""
        return self.layers(x, spatial_dim=spatial_dim, collected_outputs=collected_outputs)



"""
Wrapper for RETURNN datasets.

We make use of torch.utils.data.IterDataPipe.
"""

from __future__ import annotations

from typing import Callable, Optional, Iterable, Dict
import sys
import numpy
import torch.utils.data
from returnn.datasets.basic import Dataset as ReturnnDataset
from returnn.datasets.util.strings import str_to_numpy_array
from returnn.util.better_exchook import better_exchook

ResetCallbackT = Callable[[], None]


class ReturnnDatasetResetDefaultEpochCounterCallback:
    """
    Default for reset_callback.
    Has an internal counter for the epoch, starting at epoch 1 (RETURNN convention).
    """

    def __init__(self, dataset: ReturnnDataset):
        self.dataset = dataset
        self.epoch = 0  # next __call__ will increment, thus we start at epoch 1

    def __call__(self):
        # dataset is likely a copy of the original dataset, either in the main process or in a worker process
        self.epoch += 1
        self.dataset.init_seq_order(epoch=self.epoch)


class ReturnnDatasetResetMpSharedEpochCallback:
    """
    Can be used as reset_callback.
    """

    def __init__(self, dataset: ReturnnDataset, epoch_mp_shared: torch.multiprocessing.Value):
        self.dataset = dataset
        self.epoch_mp_shared = epoch_mp_shared

    def __call__(self):
        # dataset is likely a copy of the original dataset, either in the main process or in a worker process
        # Use epoch_mp_shared to get the current epoch correctly in worked processes
        epoch = self.epoch_mp_shared.value or None
        self.dataset.init_seq_order(epoch=epoch)


class ReturnnDatasetResetNoOpCallback:
    """
    Can be used as reset_callback.
    """

    def __call__(self):
        pass


class ReturnnDatasetIterDataPipe(torch.utils.data.IterDataPipe):
    """
    Converts a RETURNN dataset into a PyTorch IterableDataset.
    """

    def __init__(self, returnn_dataset: ReturnnDataset, *, reset_callback: Optional[ResetCallbackT] = None):
        """
        :param returnn_dataset: dataset to be wrapped
        :param reset_callback: callback function to be called when the dataset is reset, e.g. to init the epoch.
            ReturnnDatasetResetDefaultEpochCounterCallback(returnn_dataset) is the default.
        """
        self._dataset = returnn_dataset
        if not reset_callback:
            reset_callback = ReturnnDatasetResetDefaultEpochCounterCallback(returnn_dataset)
        self._reset_callback = reset_callback

    def reset(self):
        """
        This is called by PyTorch DataLoader mechanism once we create a new iterator over the DataLoader.
        This happens at the beginning of each epoch.

        (Note: The mechanism where ``reset()`` is actually called is very obfuscated in PyTorch.
        As I understand it, there is a IterDataPipe metaclass (_IterDataPipeMeta)
        which automatically registers a hook on ``__iter__`` via ``hook_iterator``.
        Deep inside the complex logic of this hook, it calls ``_set_datapipe_valid_iterator_id``
        which then calls ``reset()``.)
        """
        self._reset_callback()

    def __iter__(self) -> Iterable[Dict[str, numpy.ndarray]]:
        """
        :return: generator providing data samples in the form of a dict data_key -> data
        """
        # noinspection PyBroadException
        try:
            num_seqs = self._dataset.num_seqs
        except Exception:  # might not work for all datasets
            num_seqs = -1
        num_seqs = numpy.array(num_seqs)
        assert self._dataset.epoch is not None
        epoch = numpy.array(self._dataset.epoch)

        try:
            data_keys = self._dataset.get_data_keys()
            last_complete_frac = -1

            seq_index = 0
            while self._dataset.is_less_than_num_seqs(seq_index):
                self._dataset.load_seqs(seq_index, seq_index + 1)
                data = {data_key: self._dataset.get_data(seq_index, data_key) for data_key in data_keys}
                data["seq_tag"] = str_to_numpy_array(self._dataset.get_tag(seq_index))
                data["seq_idx"] = numpy.array(seq_index)

                # It's slightly redundant to have the following data in each entry,
                # but it's difficult to pass this back to the main proc otherwise.
                data["epoch"] = epoch
                data["num_seqs"] = num_seqs

                complete_frac = self._dataset.get_complete_frac(seq_index, allow_only_lr_suitable=True)
                if complete_frac is not None:
                    assert 0.0 <= complete_frac <= 1.0, f"complete_frac must be in [0, 1], but got {complete_frac}"
                    assert complete_frac >= last_complete_frac, (
                        "complete_frac must be monotonically increasing, "
                        f"but got {complete_frac} after {last_complete_frac}"
                    )
                else:
                    complete_frac = -1
                data["complete_frac"] = numpy.array(complete_frac, dtype=numpy.float32)
                last_complete_frac = complete_frac

                yield data
                seq_index += 1

        except Exception:
            # Report here as well, because the parent caller (PyTorch DataLoader) might first catch it
            # and only report a stripped version.
            better_exchook(*sys.exc_info(), autodebugshell=False)
            raise

    def __getitem__(self, index):
        raise Exception(f"{self.__class__.__name__}.__getitem__ not supported")


class ReturnnDatasetPerEpochMapDataPipe(torch.utils.data.MapDataPipe):
    """
    Converts a RETURNN dataset into a PyTorch map-style Dataset.
    """

    def __int__(self, returnn_dataset: ReturnnDataset, *, reset_callback: Optional[ResetCallbackT] = None):
        """
        :param returnn_dataset: dataset to be wrapped
        :param reset_callback: callback function to be called when the dataset is reset, e.g. to init the epoch.
            ReturnnDatasetResetDefaultEpochCounterCallback(returnn_dataset) is the default.
        """
        assert returnn_dataset.have_corpus_seq_idx() and returnn_dataset.have_get_corpus_seq()
        self._dataset = returnn_dataset
        if not reset_callback:
            reset_callback = ReturnnDatasetResetDefaultEpochCounterCallback(returnn_dataset)
        self._reset_callback = reset_callback

    def reset(self):
        """
        :return:
        """
        self._reset_callback()

    def __len__(self):
        """
        :return: number of data samples in the dataset
        :rtype: int
        """
        return self._dataset.num_seqs

    def __getitem__(self, index):
        """
        :param int index:
        :return: data sample in the form of a dict data_key -> data
        :rtype: dict[str, numpy.ndarray]
        """
        corpus_seq_idx = self._dataset.get_corpus_seq_idx(index)
        seq = self._dataset.get_corpus_seq(corpus_seq_idx)
        return seq.features


class ReturnnDatasetFullMapDataPipe(torch.utils.data.MapDataPipe):
    """
    Converts a RETURNN dataset into a PyTorch map-style Dataset.
    This is over the full dataset, using the default ordering.
    RETURNN-dataset-side sorting/shuffling is not supported here.
    Sorting/shuffling is intended to be done in the further PyTorch data pipeline.
    """

    def __int__(self, returnn_dataset: ReturnnDataset):
        """
        :param returnn_dataset: dataset to be wrapped
        """
        assert returnn_dataset.have_get_corpus_seq()
        self._dataset = returnn_dataset

    def __len__(self):
        """
        :return: number of data samples in the dataset
        :rtype: int
        """
        return self._dataset.get_total_num_seqs()

    def __getitem__(self, index):
        """
        :param int index:
        :return: data sample in the form of a dict data_key -> data
        :rtype: dict[str, numpy.ndarray]
        """
        seq = self._dataset.get_corpus_seq(index)
        return seq.features
